'''
nose unit tests for CFchecker.cf10
'''

import os, tempfile, shutil

from nose import tools as nt

import CFchecker
from CFchecker.cf10 import CF10Tests
from CFchecker.log import MessageLog
from CFchecker.errors import CFCheckerError

import helpers

_moduledir = os.path.dirname(__file__)

class CF10Tests_test(object):
    empty_file = 'tests_empty.nc'
    empty_file_badext = 'tests_empty.badext'
    file_noexist = 'tests_nonexistent.nc'
    faulty_file = 'tests_faulty_t2.nc'
    good_file = 'tests_good.nc'
    test_name_table = os.path.join(_moduledir, 'test_std_name_table.xml')
    test_class = CF10Tests
    #
    #   setup and teardown
    #
    @classmethod
    def setup_class(cls):
        cls.tempdir = tempfile.mkdtemp()
        helpers.create_empty_ncfile(os.path.join(cls.tempdir, cls.empty_file))
        helpers.create_empty_ncfile(os.path.join(cls.tempdir, cls.empty_file_badext))
        helpers.create_faulty_ncfile_cf10(os.path.join(cls.tempdir, cls.faulty_file))
        helpers.create_good_ncfile_cf10(os.path.join(cls.tempdir, cls.good_file))
    @classmethod
    def teardown_class(cls):
        shutil.rmtree(cls.tempdir)
    # test helper
    @classmethod
    def check_runner(self, msgs):
        '''
        runner for (almost) all _test_*_* methods
        msgs: dict with input file names as keys and expected log entries as elements of a list
        '''
        caller_name = helpers.get_caller_name()
        if not (caller_name.startswith('test_') and caller_name.endswith('_check')):
            raise Exception('check_runner does not know how to handle this caller name: %s' % caller_name)
        testid = caller_name.replace('test_', '').replace('_check', '').replace('_', '.')
        meth_name = '_test_%s' % testid.replace('.', '_')
        for inp_file in msgs.keys():
            fname = inp_file.split(os.path.sep)[-1]
            cftest = self.test_class(not_implemented_info=True, name_table_file=self.test_name_table)
            cftest._setup_run(os.path.join(self.tempdir, inp_file))
            getattr(cftest, meth_name)()
            # filter out all ('INFO', 'INIT', ...) messages for easier checking
            log = [entry for entry in cftest.get_messagelog().get_raw_log()
                   if not entry[0:2] == ('INFO', 'INIT')]
            for item in msgs[inp_file]:
                nt.assert_true(item in log, '%s: expected entry "%s" not found in log' % (fname, repr(item)))
            nt.assert_equals( len(log), len(msgs[inp_file]),
                              '%s: unexpected log entries: %s' % (fname, repr(list(set(log) - set(msgs[inp_file])))))
    #
    #   test functions
    #
    def test_init(self):
        '''CF10Tests: init'''
        # test init with list of tests to run
        cftest = self.test_class(name_table_file=self.test_name_table, tests_to_run=['2.1','2.2','2.3'])
        nt.assert_true(set(['name_table', 'tests_to_run']).issubset(set(dir(cftest))))
        nt.assert_equals(cftest.tests_to_run, ['_test_2_1', '_test_2_2', '_test_2_3'])

    def test_run(self):
        '''CF10Tests: run()'''
        cftest = self.test_class(name_table_file=self.test_name_table, tests_to_run=['2.1'])

        # test log entries with empty file and only one test selected(should only be INIT messages)
        cftest.run(os.path.join(self.tempdir, self.empty_file))
        log = cftest.get_messagelog().get_raw_log()
        nt.assert_equals(log, [('INFO', 'INIT', 'running CFchecker version %s' % (CFchecker.__version__,)),
                               ('INFO', 'INIT', 'checking compliance with convention %s' % self.test_class.cf_convention),
                               ('INFO', 'INIT', 'will only run the following tests: 2.1'),
                               ('INFO', 'INIT', 'using standard name table version: 12, ' \
                                                'last modified: 2009-07-06T12:05:02Z'),
                              ])
        # test if nonexistent file causes error
        nt.assert_raises(CFCheckerError, cftest.run, os.path.join(self.tempdir, self.file_noexist))

    def test_get_messagelog(self):
        '''CF10Tests: get_messagelog()'''
        cftest = self.test_class()
        nt.assert_true(isinstance(cftest.get_messagelog(), MessageLog))

    def test_deduce_coord_type(self):
        '''CF10Tests: _deduce_coord_type()'''
        # (units, positive, expected deduction)
        items = [(None, None, None),    # no units (None)
                 (5, 'down', None),    # no units (invalid non-string value)
                 ('foos', 'down', None), # unknown units
                 ('m', 'up', 'Z'),  # valid positive value -> Z
                 ('bar', None, 'Z'), # pressure units -> Z
                 ('level', None, 'Z'),  # deprecated level unit -> Z
                 ('degree_E', None, 'X'),   # longitide units -> X
                 ('degree_E', 'up', None),  # lon units but positive indicates Z -> nothing
                 ('degrees_north', None, 'Y'), # latitude units -> Y
                 ('hours', None, None),  # should be relative time units -> nothing
                 ('days since 2000-01-01', None, 'T'),   # time units -> T
                 ]
        for item in items:
            unit, positive, result = item
            guess = self.test_class._deduce_coord_type(unit, positive)
            nt.assert_equals(guess, result)

    def selected_test_run_check(self):
        '''CF10Tests: only run selected tests'''
        selected_testids = ['2.3', '4.3.2']
        msgs = {self.empty_file: [],
                self.faulty_file: [ ('ERROR', '2.3', 'global attribute name "_123spam" does not follow naming conventions'),
                                    ('ERROR', '2.3', 'dimension name "_spam_dim" does not follow naming conventions'),
                                    ('ERROR', '2.3', 'variable name "_spam_data" does not follow naming conventions'),
                                    ('ERROR', '2.3', 'attribute name "_spam_attrib" of variable "_spam_data" does not follow naming conventions'),
                                    ('WARNING', '2.3', 'variable names "CaseTest" and "casetest" identical except for case'),
                                    ('ERROR', '4.3.2', 'lev2: missing formula variables in file: a, b'),
                                    ('ERROR', '4.3.2', 'lev2: at least one required term could not be found in formula_terms. '
                                                       'potentially missing terms: a, ps, p0, b, ap'),
                                    ('ERROR', '4.3.2', 'lat2: standard_name "latitude" does not allow formula_terms'),
                                    ('ERROR', '4.3.2', 'lon: cannot check correctness of formula_terms because standard_name is missing'),
                                    ('ERROR', '4.3.2', 'lev3: formula_terms has invalid format'),
                                    ('ERROR', '4.3.2', 'wrong_dim_order: formula_terms attribute only allowed on coordinate variables'),
                                    ('ERROR', '4.3.2', 'lev4: detected violation of constraints for given standard_name: '
                                                       'lev4 != hya(term a) + hyb(term b) (see appendix D)'),
                                    ('ERROR', '4.3.2', 'lev5: detected violation of constraints for given standard_name: '
                                                       'lon2d(term lev) variable must be 1-dimensional'),
                                    ('ERROR', '4.3.2', 'lev6: detected violation of constraints for given standard_name: '
                                                       'ps(term p0) must be scalar or 1-dimensional with length 1'),
                                    ('WARNING', '4.3.2', 'lev7: standard_name "atmosphere_ln_pressure_coordinate" suggests '
                                                         'that there should be formula_terms, but none found'),
                                    ('ERROR', '4.3.2', 'lev8: detected violation of constraints for given standard_name: lon(term a) '
                                                       'must depend on the same single dimension as lev8'),

                                ],
                self.good_file: [],
               }
        for inp_file in msgs:
            fname = inp_file.split(os.path.sep)[-1]
            cftest = self.test_class(tests_to_run=selected_testids, name_table_file=self.test_name_table)
            cftest.run(os.path.join(self.tempdir, inp_file))
            # filter out all ('INFO', 'INIT', ...) messages for easier checking
            log = [entry for entry in cftest.get_messagelog().get_raw_log() if not entry[0:2] == ('INFO', 'INIT')]
            for item in msgs[inp_file]:
                nt.assert_true(item in log, '%s: expected entry "%s" not found in log' % (fname, repr(item)))
            nt.assert_equals( len(log), len(msgs[inp_file]),
                              '%s: unexpected log entries: %s' % (fname, repr(list(set(log) - set(msgs[inp_file])))))

    def test_2_1_check(self):
        '''CF10Tests: test 2.1'''
        msgs = {self.empty_file: [],
                self.empty_file_badext: [('ERROR', '2.1', 'file does not have .nc extension')],
                self.good_file: [],
               }
        self.check_runner(msgs)


    def test_2_2_check(self):
        '''CF10Tests: test 2.2'''
        msgs = {self.empty_file: [('NOT IMPLEMENTED', '2.2', 'not checking if CF string attributes are 1D char arrays')],
               }
        self.check_runner(msgs)

    def test_2_3_check(self):
        '''CF10Tests: test 2.3'''
        msgs = {self.faulty_file: [('ERROR', '2.3', 'global attribute name "_123spam" does not follow naming conventions'),
                                      ('ERROR', '2.3', 'dimension name "_spam_dim" does not follow naming conventions'),
                                      ('ERROR', '2.3', 'variable name "_spam_data" does not follow naming conventions'),
                                      ('ERROR', '2.3', 'attribute name "_spam_attrib" of variable "_spam_data" does not follow naming conventions'),
                                      ('WARNING', '2.3', 'variable names "CaseTest" and "casetest" identical except for case'),
                                      ],
                self.good_file: [],
                }
        self.check_runner(msgs)

    def test_2_4_check(self):
        '''CF10Tests: test 2.4'''
        msgs = { self.faulty_file: [('ERROR', '2.4', 'variable "double_dependency" has multiple dependencies on dimension "lat"'),
                                       ('WARNING', '2.4', 'variable "wrong_dim_order": dimensions other than T,Z,Y,X should be added '
                                                          'to the left according to COARDS, not to the right'),
                                       ('WARNING', '2.4', 'variable "wrong_dim_order": dimensions other than T,Z,Y,X should be added '
                                                          'to the left according to COARDS, not in between'),
                                       ('WARNING', '2.4', 'variable "wrong_dim_order": axis dimensions should be in following order: '
                                                          'T,Z,Y,X and not: Y,X,T,Z '),
                                       ('WARNING', '2.4', 'variable "no_std_name": axis dimensions should be in following order: '
                                                          'T,Z and not: Z,T '),
                                       ('WARNING', '2.4', 'variable "double_axis": axis dimensions should be in following order: '
                                                          'Y,X and not: X,X,Y ')
                                     ],
                self.good_file: [],
               }
        self.check_runner(msgs)

    def test_2_5_1_check(self):
        '''CF10Tests: test 2.5.1'''
        msgs = {self.faulty_file: [   ('WARNING', '2.5.1', 'casetest: attributes missing_value and _FillValue should have the ' \
                                                           'same value (-88888.0 <-> -99999.0)'),
                                      ('ERROR', '2.5.1', 'CaseTest: missing_value attribute and variable need to have the same data ' \
                                                         'type (int32 <-> float32)'),
                                      ('WARNING', '2.5.1', 'wrong_dim_order: _FillValue is within valid_range'),
                                      ('ERROR', '2.5.1', 'double_dependency: valid_range attribute not allowed because valid_min ' \
                                                         'and/or valid_max attributes are present.'),
                                      ('ERROR', '2.5.1', '_spam_data: _FillValue attribute and variable need to have the same data type ' \
                                                         '(int32 <-> float32)'),
                                      ('WARNING', '2.5.1', 'double_dependency: missing_value is within valid_range'),
                                     ],
                self.good_file: [],
                }
        self.check_runner(msgs)

    def test_2_6_1_check(self):
        '''CF10Tests: test 2.6.1'''
        msgs = {self.empty_file: [('ERROR', '2.6.1', 'global Conventions attribute missing (should be set to "%s")'
                                   % self.test_class.cf_convention)
                                  ],
                self.faulty_file: [('ERROR', '2.6.1', 'global Conventions attribute should be set to "%s", not "not CF-1.0"'
                                    % self.test_class.cf_convention)
                                      ],
                self.good_file: [],
               }
        self.check_runner(msgs)

    def test_2_6_2_check(self):
        '''CF10Tests: test 2.6.2'''
        msgs = {self.faulty_file: [('ERROR', '2.6.2', 'global attribute "title" has to be of type string'),
                                      ('ERROR', '2.6.2', 'global attribute "history" has to be of type string'),
                                      ('ERROR', '2.6.2', 'global attribute "institution" has to be of type string'),
                                      ('ERROR', '2.6.2', 'variable wrong_dim_order: attribute "institution" has to be of ' \
                                                         'type string'),
                                      ('ERROR', '2.6.2', 'global attribute "source" has to be of type string'),
                                      ('ERROR', '2.6.2', 'variable wrong_dim_order: attribute "source" has to be of ' \
                                                         'type string'),
                                      ('ERROR', '2.6.2', 'global attribute "references" has to be of type string'),
                                      ('ERROR', '2.6.2', 'variable wrong_dim_order: attribute "references" has to be of ' \
                                                         'type string'),
                                      ('ERROR', '2.6.2', 'global attribute "comment" has to be of type string'),
                                      ('ERROR', '2.6.2', 'variable wrong_dim_order: attribute "comment" has to be of ' \
                                                         'type string')
                                    ],
                self.good_file: [],
               }
        self.check_runner(msgs)

    def test_3_check(self):
        '''CF10Tests: test 3'''
        msgs = {self.faulty_file: [('WARNING', '3', 'variable "lev" contains neither long_name nor standard_name attribute'),
                                      ('WARNING', '3', 'variable "_spam_data" contains neither long_name nor standard_name attribute'),
                                      ('WARNING', '3', 'variable "time2" contains neither long_name nor standard_name attribute'),
                                      ('WARNING', '3', 'variable "time4" contains neither long_name nor standard_name attribute'),
                                      ('WARNING', '3', 'variable "time3" contains neither long_name nor standard_name attribute'),
                                      ('WARNING', '3', 'variable "label_var_1" contains neither long_name nor standard_name attribute'),
                                      ],
                self.good_file: [],
                }
        self.check_runner(msgs)

    def test_3_1_check(self):
        '''CF10Tests: test 3.1'''
        msgs = { self.faulty_file: [('ERROR', '3.1', 'wrong_dim_order: units are "m s-1" but should be "m s-1 m s-1" according ' \
                                                        'to standard name table (and possibly modifier definition/cell methods)'),
                                       ('ERROR', '3.1', 'lon: units "degrees_southeast" not recognized'),
                                       ('WARNING', '3.1', 'lev: usage of unit "level" is deprecated'),
                                       ('INFO', '3.1', 'no_std_name: cell_methods found but no standard_name, so not possible to ' \
                                                          'check consistency of cell_methods with units'),
                                       ('ERROR', '3.1', 'variable "time5" requires a units attribute compatible with given standard_name\'s ' \
                                                        'units ("s") (and modifier)'),
                                       ('INFO', '3.1', 'variable "time4" does not contain units attribute'),
                                       ('INFO', '3.1', 'variable "time1" does not contain units attribute'),
                                       ('ERROR', '3.1', 'time7: units "spams since 2007-01-01" not recognized, '
                                                        'see test 4.4 for more details'),
                                       ('ERROR', '3.1', 'time3: units "meters since 2000-01-01" not recognized'),
                                       ('ERROR', '3.1', 'units attribute in variable "CaseTest" is not of type string'),
                                       ('INFO', '3.1', 'variable "_spam_data" does not contain units attribute'),
                                      ],
                 self.good_file: [],
                }
        self.check_runner(msgs)

    def test_3_3_check(self):
        '''CF10Tests: test 3.3'''
        msgs = { self.faulty_file: [('ERROR', '3.3', 'CaseTest: standard_name \'spam_ratio_in_eggs\' does not exist in standard name table'),
                                      ('ERROR', '3.3', 'CaseTest: standard_name modifier \'spam_modified\' unknown'),
                                      ('ERROR', '3.3', 'time: standard_name attribute has to be of type string'),
                                      ],
                 self.good_file:     [('WARNING', '3.3', 'is_alias_name: standard_name \'surface_temperature_where_land\' is '
                                                         'an alias name. usage of the real name (surface_temperature) is recommended'),
                                     ],
                }
        self.check_runner(msgs)

    def test_3_4_check(self):
        '''CF10Tests: test 3.4'''
        msgs = { self.faulty_file: [('INFO', '3.4', 'CaseTest: standard_name of ancillary variable "wrong_dim_order" differs from ' \
                                                       'standard_name in this variable (wind_speed_of_gust <-> spam_ratio_in_eggs)'),
                                       ('ERROR', '3.4', 'CaseTest: referenced ancillary variable "does_not_exist" not found in file'),
                                       ('WARNING', '3.4', 'casetest: ancillary_variables attribute does have any usable content'),
                                       ('ERROR', '3.4', '_spam_data: ancillary_variables attribute should be of type string, not "int32"'),
                                      ],
                 self.good_file: [],
                }
        self.check_runner(msgs)

    def test_3_5_check(self):
        '''CF10Tests: test 3.5'''
        msgs = { self.faulty_file: [('ERROR', '3.5', 'casetest: flag_values attribute and variable need to have the same data type ' \
                                                        '(S <-> f)'),
                                       ('ERROR', '3.5', 'casetest: flag_meanings attribute should be a string'),
                                       ('ERROR', '3.5', 'CaseTest: flag_values and flag_meanings attributes have different lengths (3 <-> 7)'),
                                       ('WARNING', '3.5', 'CaseTest: flag_* attributes found but status_flag modifier not set'),
                                       ('ERROR', '3.5', 'wrong_dim_order: flag_meanings attribute without flag_values attribute'),
                                       ('WARNING', '3.5', 'wrong_dim_order: flag_* attributes found but status_flag modifier not set'),
                                       ('ERROR', '3.5', 'double_dependency: flag_values attribute and variable need to have the same data ' \
                                                        'type (S <-> f)'),
                                       ('ERROR', '3.5', '_spam_data: flag_values attribute and variable need to have the same data type ' \
                                                        '(i <-> f)'),
                                       ('ERROR', '3.5', '_spam_data: flag_values attribute without flag_meanings attribute'),
                                       ('WARNING', '3.5', '_spam_data: flag_* attributes found but status_flag modifier not set'),
                                       ('ERROR', '3.5', 'double_dependency: flag_values and flag_meanings attributes have '
                                                        'different lengths (45 <-> 4)'),
                                       ('ERROR', '3.5', 'double_dependency: every flag_values value has to be unique within the set of all '
                                                        'flag_values'),
                                       ('WARNING', '3.5', 'double_dependency: flag_* attributes found but status_flag modifier not set'),
                                       ('ERROR', '3.5', 'casetest: every flag_values value has to be unique within the set of all '
                                                        'flag_values'),
                                       ('WARNING', '3.5', 'casetest: flag_* attributes found but status_flag modifier not set'),
                                      ],
                 self.good_file: [],
                }
        self.check_runner(msgs)

    def test_4_check(self):
        '''CF10Tests: test 4'''
        msgs = { self.faulty_file: [('ERROR', '4', 'double_dependency: variable depends on multiple coordinate variables with '
                                                      'axis attriute "Y": lat,lat'),
                                       ('ERROR', '4', 'lon2: axis attribute inconsistent with type implied by units and/or positive '
                                                      'attributes (X <-> Y)'),
                                       ('ERROR', '4', '_spam_data: axis attributes are only allowed for coordinate variables'),
                                       ('ERROR', '4', 'double_axis: variable depends on multiple coordinate variables with axis '
                                                      'attriute "X": lon,lon2'),
                                       ('ERROR', '4', 'lat2: axis attribute contains invalid value: V'),
                                      ],
                 self.good_file: [],
                }
        self.check_runner(msgs)

    def test_4_1_check(self):
        '''CF10Tests: test 4.1'''
        msgs = { self.faulty_file: [('ERROR', '4.1', 'lat2: optional axis attribute of latitude coordinate must be set to "Y", not "V"'),
                                   ],
                 self.good_file: [('WARNING', '4.1', 'compressed_lat_lon: latitude coordinates must always explicitly include the units attribute'), ('INFO', '4.1', 'compressed_lat_lon: latitude was identified using hints in long_name, so this guess might be wrong. if so, then the warnings related to this test and variable can probably be ignored'),
                                 ],
                }
        self.check_runner(msgs)

    def test_4_2_check(self):
        '''CF10Tests: test 4.2'''
        msgs = { self.faulty_file: [('WARNING', '4.2', 'lon: longitude coordinates must have units of "degrees_east" or compatible'),
                                    ('INFO', '4.2', 'lon: longitude was identified using hints in long_name, so this guess might be '\
                                                    'wrong. if so, then the warnings related to this test and variable can probably be ignored'),
                                    ('ERROR', '4.2', 'lon3: longitude coordinates must always explicitly include the units attribute'),
                                    ('ERROR', '4.2', 'lon2: longitude coordinates must have units of "degrees_east" or compatible'),
                                    ],
                 self.good_file: [('WARNING', '4.2', 'compressed_lat_lon: longitude coordinates must always explicitly include the units attribute'),
                                  ('INFO', '4.2', 'compressed_lat_lon: longitude was identified using hints in long_name, so this guess might be wrong. if so, then the warnings related to this test and variable can probably be ignored'),
                                  ('WARNING', '4.2', 'lon: use of longitude coordinate units "degrees_east" is recommended instead of "degreesE"'),
                                  ],
                }
        self.check_runner(msgs)


    def test_4_3_check(self):
        '''CF10Tests: test 4.3'''
        msgs = { self.faulty_file: [('ERROR', '4.3', 'lev: invalid positive value "crisscross", valid attributes are only '
                                                        '"up" and "down"')
                                      ],
                 self.good_file: [],
                }
        self.check_runner(msgs)

    def test_4_3_2_check(self):
        '''CF10Tests: test 4.3.2'''
        msgs = { self.faulty_file: [('ERROR', '4.3.2', 'lev2: missing formula variables in file: a, b'),
                                    ('ERROR', '4.3.2', 'lev2: at least one required term could not be found in formula_terms. '
                                                       'potentially missing terms: a, ps, p0, b, ap'),
                                    ('ERROR', '4.3.2', 'lat2: standard_name "latitude" does not allow formula_terms'),
                                    ('ERROR', '4.3.2', 'lon: cannot check correctness of formula_terms because standard_name is missing'),
                                    ('ERROR', '4.3.2', 'lev3: formula_terms has invalid format'),
                                    ('ERROR', '4.3.2', 'wrong_dim_order: formula_terms attribute only allowed on coordinate variables'),
                                    ('ERROR', '4.3.2', 'lev4: detected violation of constraints for given standard_name: '
                                                       'lev4 != hya(term a) + hyb(term b) (see appendix D)'),
                                    ('ERROR', '4.3.2', 'lev5: detected violation of constraints for given standard_name: '
                                                       'lon2d(term lev) variable must be 1-dimensional'),
                                    ('ERROR', '4.3.2', 'lev6: detected violation of constraints for given standard_name: '
                                                       'ps(term p0) must be scalar or 1-dimensional with length 1'),
                                    ('WARNING', '4.3.2', 'lev7: standard_name "atmosphere_ln_pressure_coordinate" suggests '
                                                         'that there should be formula_terms, but none found'),
                                    ('ERROR', '4.3.2', 'lev8: detected violation of constraints for given standard_name: lon(term a) '
                                                       'must depend on the same single dimension as lev8'),
                                    ],
                 self.good_file: [],
                }
        self.check_runner(msgs)

    def test_4_4_check(self):
        '''CF10Tests: test 4.4'''
        msgs = { self.faulty_file: [('ERROR', '4.4', 'time7: units attribute contains an unknown unit: "spams"'),
                                       ('NOT IMPLEMENTED', '4.4', 'not checking if specified times are valid in the '
                                                           'given calendar'),
                                       ('WARNING', '4.4', 'time8: the unit "months" should be used with caution, '
                                                          'it is NOT a calendar month/year'),
                                       ('ERROR', '4.4', 'time6: units attribute of a time variable must have format '
                                                        '"<unit> since <reference time>"'),
                                       ('WARNING', '4.4', 'time9: The use of a reference time in the year 0 to indicate '
                                                          'climatological time is deprecated'),
                                       ('ERROR', '4.4', 'time10: unable to parse reference time of units attribute: '
                                                        'Cannot parse "01-01-01 00:00" as ISO time format'),
                                       ('WARNING', '4.4', 'time8: should not use fractional indices with unit "months" '
                                                          'as this notation is highly error-prone (it does not correspond '
                                                          'with calendar units)'),
                                      ],
                 self.good_file: [('NOT IMPLEMENTED', '4.4', 'not checking if specified times are valid in the given calendar'),],
                }
        self.check_runner(msgs)

    def test_4_4_1_check(self):
        '''CF10Tests: test 4.4.1'''
        msgs = { self.faulty_file: [('ERROR', '4.4.1', 'lev3: the attributes calendar may only be attached to '
                                                          'time coordinate variables'),
                                       ('ERROR', '4.4.1', '_spam_data: the attributes month_lengths may only be attached '
                                                          'to time coordinate variables'),
                                       ('INFO', '4.4.1', 'time: unknown calendar defined and no leap_year attribute found, it '
                                                         'is assumed that there are no leap years'),
                                       ('ERROR', '4.4.1', 'time: no or unknown calendar defined, so month_lengths attribute '
                                                          'is required'),
                                       ('INFO', '4.4.1', 'time1: not checking date/time because units attribute is missing'),
                                       ('WARNING', '4.4.1', 'time2: The time coordinate should not cross the date 1582-10-15 '
                                                             'when the default mixed Gregorian/Julian calendar is in use'),
                                       ('ERROR', '4.4.1', 'time3: the attributes calendar may only be attached to time '
                                                          'coordinate variables'),
                                       ('ERROR', '4.4.1', 'time4: leap_month attribute must be a scalar int value'),
                                       ('ERROR', '4.4.1', 'time4: month_lengths attribute has to be of type int, '
                                                          'not "float32"'),
                                       ('ERROR', '4.4.1', 'time4: leap_year attribute must be a scalar int value'),
                                       ('ERROR', '4.4.1', 'time5: leap_month attribute must be a scalar int value'),
                                       ('ERROR', '4.4.1', 'time5: month_lengths attribute must contain exactly 12 values, '
                                                          'not 6'),
                                       ('ERROR', '4.4.1', 'time5: leap_year attribute must be a scalar int value'),
                                       ('WARNING', '4.4.1', 'time6: unknown calendar defined and leap_month without '
                                                            'leap_year present, so leap_month will be ignored'),
                                       ('ERROR', '4.4.1', 'time6: leap_month value must be in range 1-12'),
                                       ('ERROR', '4.4.1', 'time6: no or unknown calendar defined, so month_lengths attribute '
                                                          'is required'),
                                       ('INFO', '4.4.1', 'time6: unknown calendar defined and no leap_year attribute found, '
                                                         'it is assumed that there are no leap years'),
                                       ('INFO', '4.4.1', 'time7: unknown or no calendar defined and leap_year present but '
                                                         'no leap_month attribute found, so february is assumed'),
                                       ('ERROR', '4.4.1', 'time7: no or unknown calendar defined, so month_lengths attribute '
                                                          'is required'),
                                       ('ERROR', '4.4.1', 'time8: no or unknown calendar defined, so month_lengths '
                                                          'attribute is required'),
                                       ],
                 self.good_file: [],
                }
        self.check_runner(msgs)

    def test_5_check(self):
        '''CF10Tests: test 5'''
        msgs = { self.faulty_file: [('ERROR', '5', 'lat: a coordinate variable must not have the _FillValue or '
                                                      'missing_value attributes'),
                                       ('ERROR', '5', 'lat2: a coordinate variable must not have the _FillValue or '
                                                      'missing_value attributes'),
                                       ('ERROR', '5', 'lon: a coordinate variable must not have the _FillValue or '
                                                      'missing_value attributes'),
                                       ('ERROR', '5', 'lat: coordinates attribute has to be of type string'),
                                       ('ERROR', '5', 'lev3: a coordinate variable must have values that are strictly '
                                                      'monotonic'),
                                       ('ERROR', '5', '_spam_data: following variables that are referenced in coordinates '
                                                      'attribute do not exist: notthere1,notthere2'),
                                       ('ERROR', '5', 'T: dimensions of auxiliary coordinate variable "lat2d" are '
                                                      'not a subset of the data variable\'s dimensions'),
                                       ('ERROR', '5', 'T: dimensions of auxiliary coordinate variable "lon2d" are '
                                                      'not a subset of the data variable\'s dimensions'),
                                       ('WARNING', '5', 'lon2d: name of a multidimensional auxiliary coordinate '
                                                        'variable should not match the name of any of its dimensions'),
                                       ('ERROR', '5', '_spam_data: dimensions of auxiliary coordinate variable '
                                                      '"label_var_1" are not a subset of the data variable\'s dimensions'),
                                       ],
                 self.good_file: [],
                }
        self.check_runner(msgs)

    def test_5_6_check(self):
        '''CF10Tests: test 5.6'''
        msgs = { self.faulty_file: [('ERROR', '5.6', '_spam_data: grid_mapping attribute has to be of type string'),
                                       ('ERROR', '5.6', 'casetest: value of grid_mapping attribute must be a single '
                                                 'variable name present in this file'),
                                       ('ERROR', '5.6', 'grid_mapping_3: grid_mapping_name "unknown_grid_mapping_name" '
                                                 'is unknown'),
                                       ('WARNING', '5.6', 'grid_mapping_1: a grid mapping variable should have 0 '
                                                   'dimensions'),
                                       ('ERROR', '5.6', 'grid_mapping_1: a grid mapping variable must have a '
                                                 'grid_mapping_name attribute'),
                                       ('ERROR', '5.6', 'grid_mapping_2: grid_mapping_name attribute has to be of type '
                                                 'string'),
                                       ('ERROR', '5.6', 'grid_mapping_4: the following attributes are not allowed for '
                                                 'the given grid_mapping_name (transverse_mercator): '
                                                 'grid_north_pole_temp,grid_north_pole_level'),
                                       ('ERROR', '5.6', 'grid_mapping_4: grid_mapping attribute latitude_of_projection_origin\'s '
                                                        'size is out of bounds'),
                                       ('ERROR', '5.6', 'grid_mapping_4: grid_mapping attribute scale_factor_at_central_meridian\'s value '
                                                 'is out of bounds'),
                                       ],
                 self.good_file: [],
                }
        self.check_runner(msgs)

    def test_6_1_check(self):
        '''CF10Tests: test 6.1'''
        msgs = { self.faulty_file: [('ERROR', '6.1', 'label_var_1: label variables must have exactly 2 '
                                                 'dimensions, not 1'),
                                       ('ERROR', '6.1', 'label_var_2: leading dimension (CDL order) must match '
                                                 'one of those of the referencing variable')
                                       ],
                 self.good_file: [],
                }
        self.check_runner(msgs)

    def test_7_1_check(self):
        '''CF10Tests: test 7.1'''
        msgs = { self.faulty_file: [('ERROR', '7.1', 'time: boundary variable casetest must have the same dimensions as '
                                                 'its associated variable, plus have a trailing dimension for the maximum number '
                                                 'of vertices in a cell'),
                                       ('WARNING', '7.1', 'casetest: boundary variables should not have the _FillValue or '
                                                   'missing_value attributes'),
                                       ('ERROR', '7.1', 'xc: bounds attribute has to be of type string'),
                                       ('ERROR', '7.1', 'yc: bounds variable "bounds_not_in_file" not found in file'),
                                       ('ERROR', '7.1', 'lat: boundary variable string_bounds must be a numeric data type'),
                                       ('WARNING', '7.1', 'lon: data is outside of boundaries for at least one cell '
                                                   '(index 1)'),
                                       ('INFO', '7.1', 'lon2: not checking if data lies within bounds of 1-sided '
                                                           'cells of 1-dimensional variables'),
                                       ],
                 self.good_file: [],
                }
        self.check_runner(msgs)

    def test_7_2_check(self):
        '''CF10Tests: test 7.2'''
        msgs = { self.faulty_file: [('ERROR', '7.2', 'casetest: cell_measures format has to be "measure: var '
                                                        '[measure: var ...]"'),
                                       ('ERROR', '7.2', '_spam_data: cell_measures attribute has to be of type string'),
                                       ('ERROR', '7.2', 'CaseTest: measure variable "does_not_exist" not found in file'),
                                       ('ERROR', '7.2', 'double_dependency: dimensions of measure variable have to be '
                                                        'subset of related variable\'s dimensions'),
                                       ('ERROR', '7.2', 'wrong_dim_order: measure variable\'s units (m/s) are not consistent '
                                                        'with given measure type (area)')
                                       ],
                 self.good_file: [],
                }
        self.check_runner(msgs)


    def test_7_3_check(self):
        '''CF10Tests: test 7.3'''
        msgs = { self.faulty_file: [('ERROR', '7.3', 'no_std_name: cell_methods dimension "lon" has to be a dimension of '
                                                         'the associated variable or a valid standard name'),
                                       ('ERROR', '7.3', 'no_std_name: cell_methods dimension "lat" has to be a dimension of '
                                                        'the associated variable or a valid standard name'),
                                       ('ERROR', '7.3', '_spam_data: cell_methods has to be of type string'),
                                       ('ERROR', '7.3', 'casetest: cell_methods could not be parsed: first piece does not end '
                                                               'with ":" (\'not\')'),
                                       ('ERROR', '7.3', 'CaseTest: dimension name "latitude" is only allowed to occur once in '
                                                        'cell_methods string'),
                                       ('ERROR', '7.3', 'double_axis: in cell_methods comment: interval unit "degrees_X" not '
                                                        'recognized'),
                                       ('ERROR', '7.3', 'double_axis: in cell_methods comment: interval length "one" not a '
                                                        'valid number'),
                                       ('ERROR', '7.3', 'wrong_dim_order: cell_methods comments could not be parsed: malformed '
                                                        'interval clause found'),
                                       ('ERROR', '7.3', 'double_dependency: cell_methods modifier "where" not allowed in CF < 1.4'),
                                       ('ERROR', '7.3', 'T: cell_methods attribute given but no content found'),
                                       ],
                 self.good_file: [],
                }
        self.check_runner(msgs)

    def test_7_4_check(self):
        '''CF10Tests: test 7.4'''
        msgs = { self.faulty_file: [('ERROR', '7.4', '_spam_data: climatology attribute has to be of type string'),
                                       ('ERROR', '7.4', 'casetest: referenced climatology variable "does_not_exist" '
                                                        'not found in file'),
                                       ('ERROR', '7.4', 'casetest: climatology attribute may only be attached to a '
                                                        'time coordinate variable'),
                                       ('ERROR', '7.4', 'xc: climatology attribute may only be attached to a time '
                                                        'coordinate variable'),
                                       ('ERROR', '7.4', 'xc: climatology variable climatology_str_bnds must have the '
                                                        'same dimension as its associated time coordinate variable, '
                                                        'and have a trailing dimension (CDL order) of size 2.'),
                                       ('ERROR', '7.4', 'climatology_str_bnds: climatology variable must be a numeric '
                                                        'data type'),
                                       ('ERROR', '7.4', 'xc: units attribute values of coordinate and climatology '
                                                        'variables must agree with each other'),
                                       ('ERROR', '7.4', 'lon3: climatology variable climatology_bnds must have the '
                                                        'same dimension as its associated time coordinate variable, '
                                                        'and have a trailing dimension (CDL order) of size 2.'),
                                       ('ERROR', '7.4', 'lon3: climatology attribute may only be attached to a time '
                                                        'coordinate variable'),
                                       ('ERROR', '7.4', 'climatology_bnds: a climatology variable must not have '
                                                        '_FillValue or missing_value attributes'),
                                       ],
                 self.good_file: [],
                }
        self.check_runner(msgs)

    def test_8_1_check(self):
        '''CF10Tests: test 8.1'''
        msgs = { self.faulty_file: [('ERROR', '8.1', '_spam_data: scale_factor and add_offset attributes must be of a '
                                                        'numeric data type'),
                                       ('ERROR', '8.1', 'casetest: scale_factor and add_offset attributes must be of same '
                                                        'numeric data type'),
                                       ('ERROR', '8.1', 'casetest: scale_factor and/or add_offset data types are different '
                                                        'from variable data type, so variable data type must be byte, short '
                                                        'or int'),
                                       ('ERROR', '8.1', 'casetest: scale_factor and/or add_offset data types are different '
                                                        'from variable data type, then attribute data type must be either '
                                                        'float or double'),
                                       ('WARNING', '8.1', 'time11: if scale_factor and/or add_offset are type float, the '
                                                          'variable should not be of type int'),
                                       ('INFO', '8.1', 'following variables seem to contain packed data: _spam_data, casetest, '
                                                       'time11. Messages from tests 4.4.1 and 7.1 concerning these '
                                                       'variables might therefore be incorrect')
                                       ],
                 self.good_file: [('INFO', '8.1', 'following variables seem to contain packed data: measurements_status, ps. '
                                                  'Messages from tests 4.4.1 and 7.1 concerning these variables might '
                                                  'therefore be incorrect')
                                 ],
                }
        self.check_runner(msgs)

    def test_8_2_check(self):
        '''CF10Tests: test 8.2'''
        msgs = { self.faulty_file: [('ERROR', '8.2', '_spam_data: compress attribute must be of type string'),
                                       ('ERROR', '8.2', 'casetest: compress attribute may only be attached to '
                                                        'coordinate variables'),
                                       ('ERROR', '8.2', 'double_axis: compress attribute may only be attached to coordinate '
                                                        'variables'),
                                       ('ERROR', '8.2', 'lat: compress attribute may only be attached to coordinate variables '
                                                        'with integer data type'),
                                       ('ERROR', '8.2', 'time11: at least one dimension name is required in compress attribute'),
                                       ('ERROR', '8.2', 'time10: following dimension names referenced in compress attribute '
                                                        'were not found in input file: does_not_exist'),
                                       ('ERROR', '8.2', 'compressed_lat_lon_time: values of variable must be in range [0;7]'),
                                       ],
                 self.good_file: [
                                 ],
                }
        self.check_runner(msgs)
