'''Helper functions for Unittests'''

import sys

from netCDF4 import Dataset as ncdf4
import numpy as np

npdt = lambda t: np.dtype(t)

def _str_to_np_arr(strlist, max_len):
    nparr = np.ma.MaskedArray(np.empty((len(strlist), max_len), dtype='|S1'), mask=True)
    for idx, entry in enumerate(strlist):
        for idy, char in enumerate(entry[:max_len]):
            nparr[idx,idy] = char
            nparr.mask[idx,idy] = False
    return nparr

def get_caller_name():
    return sys._getframe(2).f_code.co_name

def create_empty_ncfile(filename):
    ncfile = ncdf4(filename, mode='w', format='NETCDF3_CLASSIC')
    ncfile.close()
    #os.rename(filename + '.nc', filename)

def create_faulty_ncfile_cf10(filename):
    '''
    creates a test file for all tests
    should find the following problems:
    2.3:
        - invalid global attribute name '123spam'
        - invalid dimension name '_spam_dim'
        - invalid variable name '_spam_data'
        - invalid variable attribute name '_spam_attrib'
        - should NOT find invalid attribute '_FillValue' in _spam_data
        - variable names 'casetest' and 'CaseTest' equal except for case
    2.4:
        - two dependencies on 'lat' dimension in 'double_dependency' variable
        - wrong dimension order of wrong_dim_order and double_dependency variables
    2.5.1:
        - _FillValue of _spam_data has different type
        - missing_value of CaseTest has different type than var
        - double_dependency var has valid_range AND valid_min, valid_max
          AND missing_value is in valid_range
        - conflict between valid_range and _FillValue in var wrong_dim_order
        - _FillValue and missing_value in casetest have different values
    2.6.1:
        - wrong global conventions attribute "not CF-1.0"
    2.6.2:
        - wrong type of title, history, institution, source, references, comment global atts
        - wrong type of institution, source, references, comment atts in wrong_dim_order
        - title and history attributes in wrong_dim_order should NOT be found
    3:
        - _spam_data has no long_name and standard_name attributes
        - dimension variables time and lev have no long_name and standard_name attributes
    3.1:
        - lon units not known
        - lev units deprecated
        - CaseTest has units but no known standard_name (ignored by this test)
        - mismatch between wrong_dimension_order units and standard_name units
        - no_std_name has no standard_name attribute to enable check of units and cell_methods
        - units of CaseTest are no string
    3.3:
        - standard_name of time var is no string
        - standard_name of CaseTest is unknown
        - standard_name modifier of CaseTest is unknown
    3.4:
        - _spam_data has wrong type for attribute ancillary_variables
        - casetest has ancillary_variables attribute that only consists of whitespace
        - CaseTest ancillary_variables references non-existing variable "does_not_exist",
          variable without standard_name "double_dependency" (no error) and
          variable with different standard_name "wrong_dim_order"
        - variable no_std_name references variable _spam_data, both have no standard_name
          (should cause no messages)
    3.5:
        - _spam_data has flag_values attribute of different type than variable
        - _spam_data has no flag_meanings attribute
        - casetest flag_values attribute is a string
        - casetest flag_meanings attribute is no string
        - CaseTest flag_values and flag_meanings attributes have different lengths
        - double_dependency flag_values attribute is string, len of flag_meanings is different from string len
        - wrong_dim_order flag_meanings attribute without flag_values
        - wrong_dim_order flag_meanings contains invalid characters (starting from CF-1.5)
    4:
        - _spam_data data variable has axis attribute
        - dimension lat2 has invalid axis attribute
        - double_axis data variable has dependencies on two X dimensions
        - dimension variable lon2 axis attribute does not match units
        - double_dependency depends on two dimensions with axis Y
    4.3:
        - variable lev has invalid positive attribute
    4.3.2:
        - variable wrong_dim_order is no coordinate variable but has formula_terms attribute
        - coordinate variable lon has formula_terms but no standard_name
        - coordinate variable lat2 has formula_terms but wrong standard_name
        - coordinate variable lev3 has wrong formula_terms format
        - coordinate variable lev2 has wrong terms and variables in formula_terms
        - lev4 != a + b
        - lev5 term lev has to be 1-dimensional
        - lev6 term p0 has to be scalar
        - lev7 should have formula_terms
        - lev8 formula_terms references lon and lat which do not have same dimension as lev8
    4.4:
        - variable time6 has bad units format
        - variable time7 has bad time units
        - variable time3 has non-time units
        - variable time8 has months time units -> deprecated; time indices contain fractions
        - variable time9 has year 0 reference -> deprecated
        - variable time10 reference time is not in ISO format
    4.4.1:
        - variable lev3 has calendar attribute
        - variable _spam_data has month_lengths attribute
        - time has no calendar, leap_year and month_lengths
        - time1 has standard calendar, but no units attribute
        - time2 crosses date 1582-10-15
        - time3 has standard calendar, but unparsable units
        - time4 month_lengths and leap_month/leap_year have wrong type
        - time5 month_lengths and leap_month/leap_year have wrong len
        - time5 has no units attribute even though standard_name requires one
        - time6 no calendar and no month_lengths, leap_month value out of range,
          has no leap_year but leap_month
        - time7 has leap_year but no leap_month
        - time8 unknown calendar without month_lengths
    5:
        - lev3 has non-monotonic values
        - lat, lat2, lon have _FillValue/missing_value attributes
        - lat has coordinates attribute that is no string
        - _spam_data coordinates attribute rerefences nonexistent variables
    5.6:
        - _spam_data grid_mapping is no string
        - casetest grid_mapping value is no variable in this file
        - grid_mapping_1 has no grid_mapping_name and more than 0 dimensions
        - grid_mapping_2 has a non-string grid_mapping_name
        - grid_mapping_3 has an unknown grid_mapping_name
        - grid_mapping_4 has unknown grid attributes for given grid_mapping_name
        - grid_mapping_4 attribute latitude_of_projection_origin has too many elements
        - grid_mapping_4 attribute scale_factor_at_central_meridian is out of bounds
    6.1:
        - label_var_1 has wrong dimension count
        - label_var_2 leading dimension not a dimension of referencing variable
    7.1:
        - time->casetest has not the same dimensions
        - casetest has _FillValue and/or missing_value attributes
        - xc variable's bounds is no string
        - yc variable's bounds variable does not exist in file
        - string_bounds is no numerical data variable
        - lon data is out of bounds
        - lon2_bnds has unsupported shape ("1-sided" cell)
    7.2:
        - _spam_data has non-string cell_measures
        - casetest has invalid cell_measures format
        - CaseTest references measure variable that does not exist
        - double_dependency references casetest whose dimensions are no subset
        - wrong_dim_order references area sizes with wrong units attribute
    7.3:
        - no_std_name has cell_methods with wrong dimensions
        - _spam_data has cell_methods that are not a string
        - casetest has invalid cell_methods format
        - CaseTest cell_methods references same dimension twice
        - double_axis cell_methods comment has unknown units and non-number interval len
        - wrong_dim_order cell_methods comment has malformed interval clause
    7.4:
        - _spam_data has non-string climatology attribute
        - casetest climatology attribute does not exist in file and casetest is no dim var
        - xc is no time variable (identified by units) but has climatology attribute
        - xc and climatology_str_bnds do not have the same dimension
        - xc and climatology_str_bnds do not have same units
        - climatology_str_bnds is not a numeric data type
        - lon3 is no time variable (identified by standard_name)
        - lon3 and climatology_bnds dimensions do not match
        - climatology_bnds has _FillValue
    8.1:
        - _spam_data add_offset and scale_factor attributes are non-numeric
        - casetest add_offset and scale_factor attributes are not of same numeric type
        - casetest vartype and scale_factor attribute type are not the same and
          scale_factor is not of type float or double
        - time11 type is int and scale_factor type is float
    8.2:
        - _spam_data has non-string compress attribute
        - casetest has compress attribute but is neither coordinate variable nor of type int
        - double_axis has compress attribute but is no coordinate variable
        - lat variable has compress attribute but is not of type int
        - time11 has compress attribute string containing only spaces
        - time10 compress attribute references dimension that does not exist
        - compressed_lat_lon_time has valid compress attribute but out-of-range indices
    '''
    ncfile = ncdf4(filename, mode='w', format='NETCDF3_CLASSIC')
    try:
        #
        # create global attributes
        #
        ncfile.setncatts({ '_123spam': 'this attribute name is invalid',
                                   'Conventions': 'not CF-1.0',
                                   'title': np.array([-1], dtype=np.float32),
                                   'history': np.array([-2], dtype=np.float32),
                                   'institution': np.array([-3], dtype=np.float32),
                                   'source': np.array([-4], dtype=np.float32),
                                   'references': np.array([-5], dtype=np.float32),
                                   'comment': np.array([-6], dtype=np.float32),
                                   })
        #
        # create dimensions
        ncfile.createDimension('_spam_dim', 1)
        ncfile.createDimension('lat', 2)
        ncfile.createDimension('lon', 2)
        ncfile.createDimension('lev', 2)
        ncfile.createDimension('lat2', 2)
        ncfile.createDimension('lon2', 2)
        ncfile.createDimension('lon3', 2)
        ncfile.createDimension('lev2', 2)
        ncfile.createDimension('lev3', 4)
        ncfile.createDimension('lev4', 4)
        ncfile.createDimension('lev5', 4)
        ncfile.createDimension('lev6', 4)
        ncfile.createDimension('lev7', 4)
        ncfile.createDimension('lev8', 4)
        ncfile.createDimension('time', 2)
        ncfile.createDimension('time1', 2)
        ncfile.createDimension('time2', 25)
        ncfile.createDimension('time3', 2)
        ncfile.createDimension('time4', 2)
        ncfile.createDimension('time5', 2)
        ncfile.createDimension('time6', 2)
        ncfile.createDimension('time7', 2)
        ncfile.createDimension('time8', 2)
        ncfile.createDimension('time9', 2)
        ncfile.createDimension('time10', 2)
        ncfile.createDimension('time11', 2)
        ncfile.createDimension('bnds', 2)
        ncfile.createDimension('xc', 2)
        ncfile.createDimension('yc', 2)
        ncfile.createDimension('lon2d', 2)
        ncfile.createDimension('max_label_len', 5)
        ncfile.createDimension('compressed_lat_lon_time', 3)
        #
        # create and fill dimension variables
        newvar = ncfile.createVariable('compressed_lat_lon_time', np.int32, dimensions=('compressed_lat_lon_time',))
        newvar[:] = np.array([1,2,12], dtype=np.int32)
        newvar.setncatts({ 'compress': 'lat lon time',
                                        'long_name': 'compress indices for lat, lon, time',
                                        })

        newvar = ncfile.createVariable('xc', np.float32, dimensions=('xc',))
        newvar[:] = np.array([1,2], dtype=np.float32)
        newvar.setncatts({ 'units': 'm',
                                        'long_name': 'x-coordinate in Cartesian system',
                                        'bounds': np.array([1], dtype=np.int32),
                                        'climatology': 'climatology_str_bnds',
                                        })

        newvar = ncfile.createVariable('yc', np.float32, dimensions=('yc',))
        newvar[:] = np.array([1,2], dtype=np.float32)
        newvar.setncatts({ 'units': 'm',
                                       'long_name': 'y-coordinate in Cartesian system',
                                       'bounds': 'bounds_not_in_file',
                                       })

        newvar = ncfile.createVariable('lat', np.float32, dimensions=('lat',))
        newvar[:] = np.array([1,2], dtype=np.float32)
        newvar.setncatts({ 'axis': 'Y',
                                       'standard_name': 'latitude',
                                       'units': 'degrees_north',
                                       '_FillValue': np.array(404.0, dtype=np.float32),
                                       'coordinates': np.array(104.0, dtype=np.float32),
                                       'bounds': 'string_bounds',
                                       'compress': 'lat',
                                       })

        newvar = ncfile.createVariable('lat2', np.float32, dimensions=('lat2',))
        newvar[:] = np.array([5,6], dtype=np.float32)
        newvar.setncatts({ 'axis': 'V',
                                       'standard_name': 'latitude',
                                       'units': 'degrees_north',
                                       'formula_terms': 'a: a b: b',
                                       'missing_value': np.array(404.0, dtype=np.float32),
                                       })

        newvar = ncfile.createVariable('lon', np.float32, dimensions=('lon',))
        newvar[:] = np.array([3,4], dtype=np.float32)
        newvar.setncatts( { 'axis': 'X',
                                       'long_name': 'longitude',
                                       'units': 'degrees_southeast',
                                       'formula_terms': 'a: a b: b',
                                       '_FillValue': np.array(404.0, dtype=np.float32),
                                       'missing_value': np.array(404.0, dtype=np.float32),
                                       'bounds': 'lon_bnds',
                                       })

        newvar = ncfile.createVariable('lon2', np.float32, dimensions=('lon2',))
        newvar[:] = np.array([8,9], dtype=np.float32)
        newvar.setncatts( { 'axis': 'X',
                                        'standard_name': 'longitude',
                                        'units': 'degrees_N',
                                        'bounds': 'lon2_bnds',
                                        })

        newvar = ncfile.createVariable('lon3', np.float32, dimensions=('lon3',))
        newvar[:] = np.array([8,9], dtype=np.float32)
        newvar.setncatts( { 'axis': 'X',
                                       'standard_name': 'longitude',
                                       'climatology': 'climatology_bnds',
                                       })

        newvar = ncfile.createVariable('lev', np.float32, dimensions=('lev',))
        newvar[:] = np.array([5,6], dtype=np.float32)
        newvar.setncatts( { 'axis': 'Z',
                                       'units': 'level',
                                       'positive': 'crisscross',
                                       })

        newvar = ncfile.createVariable('lev2', np.float32, dimensions=('lev2',))
        newvar[:] = np.array([9,10], dtype=np.float32)
        newvar.setncatts( { 'units': '1',
                                       'positive': 'DOWN',
                                       'long_name': 'another level coordinate axis',
                                       'standard_name': 'atmosphere_hybrid_sigma_pressure_coordinate',
                                       'formula_terms': 'abc: a def: b',
                                       })

        newvar = ncfile.createVariable('lev3', np.float32, dimensions=('lev3',))
        newvar[:] = np.array([5,6,1,2], dtype=np.float32)
        newvar.setncatts( { 'axis': 'Z',
                                        'units': '1',
                                        'standard_name': 'atmosphere_hybrid_sigma_pressure_coordinate',
                                        'formula_terms': 'a:a b:b',
                                        'calendar': 'standard',
                                        })

        newvar = ncfile.createVariable('lev4', np.float32, dimensions=('lev4',))
        newvar[:] = np.array([1,2,3,4], dtype=np.float32)
        newvar.setncatts( { 'axis': 'Z',
                                        'units': '1',
                                        'standard_name': 'atmosphere_hybrid_sigma_pressure_coordinate',
                                        'formula_terms': 'a: hya b: hyb ps: ps p0: p0',
                                        })

        newvar = ncfile.createVariable('lev5', np.float32, dimensions=('lev5',))
        newvar[:] = np.array([1,2,3,4], dtype=np.float32)
        newvar.setncatts( { 'axis': 'Z',
                                        'units': '1',
                                        'standard_name': 'atmosphere_ln_pressure_coordinate',
                                        'formula_terms': 'p0: p0 lev: lon2d',
                                        })

        newvar = ncfile.createVariable('lev6', np.float32, dimensions=('lev6',))
        newvar[:] = np.array([1,2,3,4], dtype=np.float32)
        newvar.setncatts( { 'axis': 'Z',
                                        'units': '1',
                                        'standard_name': 'atmosphere_ln_pressure_coordinate',
                                        'formula_terms': 'p0: ps lev: hya',
                                        })

        newvar = ncfile.createVariable('lev7', np.float32, dimensions=('lev7',))
        newvar[:] = np.array([1,2,3,4], dtype=np.float32)
        newvar.setncatts( { 'axis': 'Z',
                                        'units': '1',
                                        'standard_name': 'atmosphere_ln_pressure_coordinate',
                                        })

        newvar = ncfile.createVariable('lev8', np.float32, dimensions=('lev8',))
        newvar[:] = np.array([1,2,3,4], dtype=np.float32)
        newvar.setncatts( { 'axis': 'Z',
                                        'units': '1',
                                        'standard_name': 'atmosphere_hybrid_sigma_pressure_coordinate',
                                        'formula_terms': 'a: lon b: lat ps: ps p0: p0',
                                        })

        newvar = ncfile.createVariable('time', np.int32, dimensions=('time',))
        newvar[:] = np.array([0,1], dtype=np.int32)
        newvar.setncatts( { 'axis': 'T',
                                        'bounds': 'casetest',
                                        'units': 'days since 2000-01-01',
                                        'standard_name': np.array([-7], dtype=np.float32),
                                        })

        # used to test _find_time_vars START
        newvar = ncfile.createVariable('time1', np.int32, dimensions=('time1',))
        newvar[:] = np.array([0,1], dtype=np.int32)
        newvar.setncatts( { 'long_name': 'time',
                                        'calendar': 'standard',
                                        })

        newvar = ncfile.createVariable('time2', np.int32, dimensions=('time2',))
        newvar[:] = np.array(list(xrange(0,25)), dtype=np.int32)
        newvar.setncatts( { 'units': 'days since 1582-10-10 12:00:00.0',
                                        'calendar': 'standard',
                                        })

        newvar = ncfile.createVariable('time3', np.int32, dimensions=('time3',))
        newvar[:] = np.array([0,1], dtype=np.int32)
        newvar.setncatts( { 'units': 'meters since 2000-01-01',
                                        'calendar': 'standard',
                                        })

        newvar = ncfile.createVariable('time4', np.int32, dimensions=('time4',))
        newvar[:] = np.array([0,1], dtype=np.int32)
        newvar.setncatts( { 'axis': 'T',
                                        'month_lengths': np.array([1.2,4.5,4.5], dtype=np.float32),
                                        'leap_month': np.array([1], dtype=np.float32),
                                        'leap_year': np.array([1], dtype=np.float32),
                                        })

        newvar = ncfile.createVariable('time5', np.int32, dimensions=('time5',))
        newvar[:] = np.array([0,1], dtype=np.int32)
        newvar.setncatts( { 'standard_name': 'time',
                                        'month_lengths': np.array([1,2,4,5,4,5], dtype=np.int32),
                                        'leap_month': np.array([1,2,3], dtype=np.int32),
                                        'leap_year': np.array([1,2,3], dtype=np.int32),
                                        })
        # used to test _find_time_vars END

        newvar = ncfile.createVariable('time6', np.int32, dimensions=('time6',))
        newvar[:] = np.array([0,1], dtype=np.int32)
        newvar.setncatts( { 'standard_name': 'time',
                                        'units': 'days',
                                        'leap_month': 13,
                                        })

        newvar = ncfile.createVariable('time7', np.int32, dimensions=('time7',))
        newvar[:] = np.array([0,1], dtype=np.int32)
        newvar.setncatts( { 'standard_name': 'time',
                                        'units': 'spams since 2007-01-01',
                                        'leap_year': np.array([2000], dtype=np.int32),
                                        })

        newvar = ncfile.createVariable('time8', np.float32, dimensions=('time8',))
        newvar[:] = np.array([0.5,1], dtype=np.float32)
        newvar.setncatts( { 'standard_name': 'time',
                                        'units': 'months since 2000-01-01',
                                        'calendar': 'ham',
                                        'leap_year': np.array([2000], dtype=np.int32),
                                        'leap_month': np.array([2], dtype=np.int32),
                                        })

        newvar = ncfile.createVariable('time9', np.int32, dimensions=('time9',))
        newvar[:] = np.array([0,1], dtype=np.int32)
        newvar.setncatts( { 'standard_name': 'time',
                                        'units': 'days since 0000-01-01 00:00',
                                        'calendar': 'julian',
                                        })

        newvar = ncfile.createVariable('time10', np.int32, dimensions=('time10',))
        newvar[:] = np.array([0,1], dtype=np.int32)
        newvar.setncatts( { 'standard_name': 'time',
                                        'units': 'days since 01-01-01 00:00',
                                        'calendar': 'all_leap',
                                        'compress': 'time does_not_exist',
                                        })

        newvar = ncfile.createVariable('time11', np.int32, dimensions=('time11',))
        newvar[:] = np.array([0,1], dtype=np.int32)
        newvar.setncatts( { 'standard_name': 'time',
                                        'climatology': 'climatology_bnds',
                                        'calendar': 'all_leap',
                                        'units': 'days since 2000-01-01',
                                        'scale_factor': np.array(2.5,  dtype=np.float32),
                                        'compress': '   ',
                                        })
        #
        # create data variables and variable attributes
        #
        # lon2d, lat2d
        newvar = ncfile.createVariable('lon2d', np.float32, dimensions=('lon2d','xc'))
        newvar[:] = np.array([[1,2],[3,4]], dtype=np.float32)
        newvar.setncatts( { 'units': 'degrees_east',
                                       'long_name': 'longitude',
                                       'standard_name': 'longitude',
                                       })

        newvar = ncfile.createVariable('lat2d', np.float32, dimensions=('yc','xc'))
        newvar[:] = np.array([[1,2],[3,4]], dtype=np.float32)
        newvar.setncatts( { 'units': 'degrees_north',
                                        'long_name': 'latitude',
                                        'standard_name': 'latitude',
                                        })
        # _spam_data
        newvar = ncfile.createVariable('_spam_data', np.float32, dimensions=('_spam_dim',))
        newvar[:] = np.array([42.0], dtype=np.float32)
        newvar.setncatts( { '_FillValue': np.array([-99999], dtype=np.int32),
                                       '_spam_attrib': 'this attribute name is invalid as well',
                                       'flag_values': np.array([-9, -8, -7], dtype=np.int32),
                                       'ancillary_variables': np.array([-9, -8, -7], dtype=np.int32),
                                       'axis': 'XYZ',
                                       'month_lengths': np.array([1,1,1,1,1,1,1,1,1,1,1,1], dtype=np.int32),
                                       'coordinates': 'CaseTest notthere1 notthere2 casetest label_var_1',
                                       'grid_mapping': np.array(1, dtype=np.float32),
                                       'cell_measures': 1,
                                       'cell_methods': 1,
                                       'climatology': 1,
                                       'scale_factor': 'should be numtype',
                                       'add_offset': 'should be numtype',
                                       'compress': 1,
                                       })
        # casetest
        newvar = ncfile.createVariable('casetest', np.float32, dimensions=('_spam_dim',))
        newvar[:] = np.array([42.23], dtype=np.float32)
        newvar.setncatts( { 'missing_value': np.array([-88888], dtype=np.float32),
                                        '_FillValue': np.array([-99999], dtype=np.float32),
                                        'flag_values': 'this should not be a string',
                                        'flag_meanings': np.array([-9, -10, -11], dtype=np.int32),
                                        'ancillary_variables': '      ',
                                        'grid_mapping': 'not a variable in this file',
                                        'cell_measures': 'not a valid cell_measures format',
                                        'cell_methods': 'not a valid cell_methods format',
                                        'climatology': 'does_not_exist',
                                        'scale_factor': 1,
                                        'add_offset': 42.5,
                                        'compress': 'lat',
                                        })
        # CaseTest
        newvar = ncfile.createVariable('CaseTest', np.float32, dimensions=('_spam_dim',))
        newvar[:] = np.array([42.5], dtype=np.float32)
        newvar.setncatts( { 'missing_value': np.array([-88888], dtype=np.int32),
                                        'standard_name': 'spam_ratio_in_eggs spam_modified',
                                        'units': 1,
                                        'flag_values': np.array([-9, -8, -7], dtype=np.float32),
                                        'flag_meanings': 'this attribute should only have three elements',
                                        'ancillary_variables': 'does_not_exist double_dependency wrong_dim_order',
                                        'grid_mapping': 'grid_mapping_1',
                                        'cell_measures': 'area: does_not_exist',
                                        'cell_methods': 'latitude: variance latitude: sum',
                                        })
        # grid_mapping_1
        newvar = ncfile.createVariable('grid_mapping_1', np.float32, dimensions= ('_spam_dim',))
        newvar[:] = np.array([423], dtype=np.float32)
        # double_dependency
        newvar = ncfile.createVariable('double_dependency', np.float32, dimensions=('lat', 'lon', 'lat'))
        newvar[:] = np.array([[[1,2],[3,4]],[[5,6],[7,8]]], dtype=np.float32)
        newvar.setncatts( { 'valid_range': np.array([0.0, 10.0], dtype=np.float32),
                                        'valid_min': np.array([0.0], dtype=np.float32),
                                        'valid_max': np.array([10.0], dtype=np.float32),
                                        'missing_value': np.array([5], dtype=np.float32),
                                        'cell_methods': 'time: sum where sea_ice within years time: mean over years' ,
                                        'flag_values': 'this string is far too long for flag_meanings',
                                        'flag_meanings': 'lengths will not match',
                                        'grid_mapping': 'grid_mapping_2',
                                        'cell_measures': 'area: casetest',
                                        })
        # grid_mapping_2
        newvar = ncfile.createVariable('grid_mapping_2', 'S1', dimensions=())
        newvar[:] = _str_to_np_arr(['x'], 1)
        newvar.setncatts( { 'grid_mapping_name': np.array(423, dtype=np.float32),
                                        })
        # climatology_str_bnds
        newvar = ncfile.createVariable('climatology_str_bnds', 'S1', dimensions=('time', 'bnds'))
        newvar[:] = _str_to_np_arr(['AB', 'CD'], len(ncfile.dimensions['bnds']))
        newvar.setncatts( { 'long_name': 'climatology bounds',
                                        'units': '1',
                                        })
        # climatology_bnds
        newvar = ncfile.createVariable('climatology_bnds', np.float32, dimensions=('time11', 'bnds'))
        newvar[:] = np.array(list(xrange(2**2)), dtype=np.float32).reshape(2,2)
        newvar.setncatts( { 'long_name': 'climatology bounds',
                                        '_FillValue': np.array(999.9999, dtype=np.float32),
                                        })

        # double_axis
        newvar = ncfile.createVariable('double_axis', np.int32, dimensions=('lon', 'lon2', 'lat'))
        newvar[:] = np.array(list(xrange(2**3)), dtype=np.int32).reshape(2,2,2)
        newvar.setncatts( { 'units': 'm/s',
                                        'standard_name': 'wind_speed_of_gust',
                                        'grid_mapping': 'grid_mapping_3',
                                        'cell_methods': 'lat: maximum lon: maximum (interval: 5.0 degrees_X '
                                                               'interval: one degrees_E interval: 5 seconds '
                                                               'comment: some errors in this comment section)',
                                        'compress': 'lat',
                                        })
        # grid_mapping_3
        newvar = ncfile.createVariable('grid_mapping_3', np.float32, dimensions=())
        newvar[:] = np.array([1],dtype=np.float32)
        newvar.setncatts( { 'grid_mapping_name': 'unknown_grid_mapping_name',
                                        })
        # wrong_dim_order
        newvar = ncfile.createVariable('wrong_dim_order', np.float32,
                                        dimensions=('_spam_dim', 'lat', 'bnds', 'lon', 'time', 'lev'))
        newvar[:] = np.array(list(xrange(0,2**5)), dtype=np.float32).reshape(1,2,2,2,2,2)
        newvar.setncatts( { 'valid_range': np.array([0.0, 300.0], dtype=np.float32),
                                        '_FillValue': np.array([300.0], dtype=np.float32),
                                        'title': np.array([-6], dtype=np.float32),
                                        'history': np.array([-8], dtype=np.float32),
                                        'institution': np.array([5555], dtype=np.float32),
                                        'source': np.array([66666], dtype=np.float32),
                                        'references': np.array([77777], dtype=np.float32),
                                        'comment': np.array([88888], dtype=np.float32),
                                        'long_name': 'wrong dimension order',
                                        'units': 'm s-1',
                                        'cell_methods': 'lon: sum lat: variance (interval: 5.0 comment: units missing in interval)',
                                        'standard_name': 'wind_speed_of_gust',
                                        'flag_meanings': 'flag#values mis$in~g in* this v=ar',
                                        'formula_terms': 'a: a b: b',
                                        'cell_measures': 'area: area_sizes_wrong_u',
                                        })
        # area_sizes_wrong_u
        newvar = ncfile.createVariable('area_sizes_wrong_u', np.float32, dimensions=('lat', 'lon'))
        newvar[:] = np.array(list(range(0,2**2)), dtype=np.float32).reshape(2,2)
        newvar.setncatts( { 'long_name': 'area sizes',
                                        'units': 'm/s',
                                        })
        # T
        newvar = ncfile.createVariable('T', np.float32, dimensions=('lev', 'lat', 'xc'))
        newvar[:] = np.array(list(range(0,2**3)), dtype=np.float32).reshape(2,2,2)
        newvar.setncatts( { 'long_name': 'temperature',
                                        'standard_name': 'air_potential_temperature',
                                        'units': 'K',
                                        'coordinates': 'lon2d lat2d label_var_2',
                                        'grid_mapping': 'grid_mapping_4',
                                        'cell_methods': ' ',
                                        })
        # hya
        newvar = ncfile.createVariable('hya', np.float32, dimensions=('lev4',))
        newvar[:] = [0.5,0.6,0.7,0.8]
        newvar.setncatts( { 'long_name': 'atmosphere hybrid sigma a coordinate',
                                        'units': '1',
                                        })
        # hyb
        newvar = ncfile.createVariable('hyb', np.float32, dimensions=('lev4',))
        newvar[:] = [0.1,0.2,0.3,0.4]
        newvar.setncatts( { 'long_name': 'atmosphere hybrid sigma b coordinate',
                                        'units': '1',
                                        })
        # p0
        newvar = ncfile.createVariable('p0', np.float32, dimensions=())
        newvar[:] = 1000
        newvar.setncatts( { 'long_name': 'reference pressure',
                                        'units': 'hPa',
                                        })
        # ps
        newvar = ncfile.createVariable('ps', np.float32, dimensions=('time', 'lat', 'lon'))
        newvar[:] = np.array([1024.0]* 2**3, dtype=np.float32).reshape(2,2,2)
        newvar.setncatts( { 'long_name': 'surface pressure',
                                        'units': 'hPa',
                                        })
        # grid_mapping_4
        newvar = ncfile.createVariable('grid_mapping_4', 'S1', dimensions=())
        newvar.setncatts( { 'grid_mapping_name': 'transverse_mercator',
                                        'scale_factor_at_central_meridian': np.array(-80.0, dtype=np.float32),
                                        'grid_north_pole_level': np.array(120.0, dtype=np.float32),
                                        'grid_north_pole_temp': np.array(-70.0, dtype=np.float32),
                                        'latitude_of_projection_origin': np.array([-70.0,30], dtype=np.float32),
                                        })
        # no_std_name
        newvar = ncfile.createVariable('no_std_name', np.float32, dimensions=('lev2','time',))
        newvar[:] = np.array([[0.2,0.3],[0.6,0.7]], dtype=np.float32)
        newvar.setncatts( { 'cell_methods': 'lon: sum lat: variance',
                                        'long_name': 'cell_methods and units but no standard_name',
                                        'units': 'm s-1',
                                        'ancillary_variables': '_spam_data',
                                        })
        # label_var_1
        newvar = ncfile.createVariable('label_var_1', 'S1', dimensions=('lev2',))
        newvar[:] = _str_to_np_arr(['XY'], len(ncfile.dimensions['lev2']))
        # label_var_2
        newvar = ncfile.createVariable('label_var_2', 'S1', dimensions=('lon','max_label_len'))
        newvar[:] = _str_to_np_arr(['XY', 'ZAB'], len(ncfile.dimensions['max_label_len']))
        # string_bounds
        newvar = ncfile.createVariable('string_bounds', 'S1', dimensions=('lat','bnds'))
        newvar[:] = _str_to_np_arr(['XY','AB'], len(ncfile.dimensions['bnds']))
        # lon_bnds
        newvar = ncfile.createVariable('lon_bnds', np.float32, dimensions=('lon','bnds'))
        newvar[:] = np.array([[2,3.5],[4.5,6]], dtype=np.float32)
        # lon2_bnds
        newvar = ncfile.createVariable('lon2_bnds', np.float32, dimensions=('lon2','_spam_dim'))
        newvar[:] = np.array([[3.5],[4.5]], dtype=np.float32)
    finally:
        ncfile.close()

def create_good_ncfile_cf10(filename):
    '''
    creates a test file that should be CF-1.0 compliant
    contains:
        - bounds for utils.get_bounds test
    general error handling:
        - deal with 0-sized (unlimited) dimensions
        - correctly identify _FillValue == missing_value when value is NaN
    2.3:
        - respecting naming conventions
        - no var names equal except for case
    2.4:
        - dimensions of variables have all different names
        - dimension order is TZYX...
    2.5.1:
        - valid_min and valid_max in variable good_dim_order, but no valid_range
        - _FillValue same type as variable good_dim_order
        - no missing_value attribute
    2.6.1:
        - correct global conventions attribute "CF-1.0"
    2.6.2:
        - title, history, institution, source, references,
          comment global attributes (strings)
    3.1:
        - correct units/standard_name for good_dim_order and measurements_status
    3.3:
        - good standard_name and modifiers for good_dim_order and measurements_status
        - variable is_alias_name has a standard_name that is an alias
    3.4:
        - variable good_dim_order references ancillary variable measurements_status
          which has the same standard_name
    3.5:
        - measurements_status has valid flag_values and flag_meanings attributes
        - measurements_status_2 has valid flag_values and flag_meanings strings
    4.3:
        - lev variable has valid positive value
    4.3.2:
        - lev variable has valid formula_terms attribute (as far as the test is concerned)
        - lev variable has valid formula_terms attribute without p0
    4.4:
        - time variable units has correct format
    4.4.1:
        - time variable has standard calendar
    5:
        - handles multidim coordinate variables (also with labels) correctly
    6.1:
        - recognizes lev_name as label variable
    7.1:
        - time has valid bounds
    7.2:
        - good_dim_order has valid cell_measures
    8.1:
        - measurement_status has valid scale_factor attribute
        - ps has valid scale_factor and add_offset attributes
    '''
    ncfile = ncdf4(filename, mode='w', format='NETCDF3_CLASSIC')
    try:
        #
        # create global attributes
        #
        ncfile.setncatts( { 'Conventions': 'CF-1.0',
                                    'title': 'CF-1.0 compliant test file',
                                    'history': 'this has always been a test file',
                                    'institution': 'FZJ',
                                    'source': 'unit test helpers',
                                    'references': 'none',
                                    'comment': 'this file has been created as a CF-1.0 compliant test file',
                                    })
        #
        # create dimensions
        ncfile.createDimension('lat', 2)
        ncfile.createDimension('lon', 2)
        ncfile.createDimension('lev', 2)
        ncfile.createDimension('lev2', 2)
        ncfile.createDimension('time', 2)
        ncfile.createDimension('bnds', 2)
        ncfile.createDimension('xc', 2)
        ncfile.createDimension('yc', 2)
        ncfile.createDimension('max_label_len', 6)
        ncfile.createDimension('compressed_lat_lon', 2)
        ncfile.createDimension('zero_sized_unlimited_dim', None)
        #
        # create and fill dimension variables
        # add axis attriburtes
        # lat, lon, lev, time
        newvar = ncfile.createVariable('xc', np.float32, dimensions=('xc',))
        newvar[:] = np.array([1,2], dtype=np.float32)
        newvar.setncatts( { 'units': 'm',
                                        'long_name': 'x-coordinate in Cartesian system',
                                        })

        newvar = ncfile.createVariable('yc', np.float32, dimensions=('yc',))
        newvar[:] = np.array([1,2], dtype=np.float32)
        newvar.setncatts( { 'units': 'm',
                                        'long_name': 'y-coordinate in Cartesian system',
                                        })

        newvar = ncfile.createVariable('compressed_lat_lon', np.int32, dimensions=('compressed_lat_lon',))
        newvar[:] = np.array([1,2], dtype=np.int32)
        newvar.setncatts( { 'long_name': 'compressed latitude and longitude indices',
                                        'compress': 'lat lon',
                                        })

        newvar = ncfile.createVariable('lat', np.float32, dimensions=('lat',))
        newvar[:] = np.array([1,2], dtype=np.float32)
        newvar.setncatts( { 'axis': 'Y',
                                        'long_name': 'latitude',
                                        'standard_name': 'latitude',
                                        'units': 'degrees_north',
                                        })

        newvar = ncfile.createVariable('lon', np.float32, dimensions=('lon',))
        newvar[:] = np.array([3,4], dtype=np.float32)
        newvar.setncatts( { 'axis': 'X',
                                        'long_name': 'longitude',
                                        'standard_name': 'longitude',
                                        'units': 'degreesE',
                                        })

        newvar = ncfile.createVariable('lev', np.float32, dimensions=('lev',))
        newvar[:] = np.array([0.7,0.61], dtype=np.float32)
        newvar.setncatts( { 'axis': 'Z',
                                        'long_name': 'hybrid sigma pressure coordinate',
                                        'standard_name': 'atmosphere_hybrid_sigma_pressure_coordinate',
                                        'units': '1',
                                        'positive': 'UP',
                                        'formula_terms': 'p0: p0 ap: ap b: b ps: ps',
                                        'bounds': 'lev_bnds',
                                        })

        newvar = ncfile.createVariable('lev2', np.float32, dimensions=('lev2',))
        newvar[:] = np.array([0.7,0.61], dtype=np.float32)
        newvar.setncatts( { 'axis': 'Z',
                                        'long_name': 'hybrid sigma pressure coordinate',
                                        'standard_name': 'atmosphere_hybrid_sigma_pressure_coordinate',
                                        'units': '1',
                                        'positive': 'UP',
                                        'formula_terms': 'ap: ap2 b: b2 ps: ps',
                                        })

        newvar = ncfile.createVariable('time', np.int32, dimensions=('time',))
        newvar[:] = np.array([1,3], dtype=np.int32)
        newvar.setncatts( { 'axis': 'T',
                                        'long_name': 'time',
                                        'standard_name': 'time',
                                        'units': 'days since 2000-01-01',
                                        'calendar': 'gregorian',
                                        'bounds': 'time_bnds',
                                        })

        newvar = ncfile.createVariable('zero_sized_unlimited_dim', np.int32, dimensions=('zero_sized_unlimited_dim',))
        # do NOT add any data because it's a zero-sized dim!
        newvar.setncatts({'long_name': 'zero-sized dimension variable',
                                       'standard_name': 'time',
                                       'units': 'days since 2000-01-01',
                                       'calendar': 'gregorian',
                                        })

        #
        # create data variables and variable attributes
        #
        # lev_bnds
        newvar = ncfile.createVariable('lev_bnds', np.float32, dimensions=('lev','bnds'))
        newvar[:] = np.array([[0.65,0.75],[0.60,0.62]], dtype=np.float32)

        # lon2d, lat2d
        newvar = ncfile.createVariable('lon2d', np.float32, dimensions=('yc','xc'))
        newvar[:] = np.array([[1,2],[3,4]], dtype=np.float32)
        newvar.setncatts( { 'units': 'degrees_east',
                                        'long_name': 'longitude',
                                        'standard_name': 'longitude',
                                        })

        newvar = ncfile.createVariable('lat2d', np.float32, dimensions=('yc','xc'))
        newvar[:] = np.array([[1,2],[3,4]], dtype=np.float32)
        newvar.setncatts( { 'units': 'degrees_north',
                                        'long_name': 'latitude',
                                        'standard_name': 'latitude',
                                        })
        # time_bnds
        newvar = ncfile.createVariable('time_bnds', np.int32, dimensions=('time', 'bnds'))
        newvar[:] = np.array([[0,2],[2,4]], dtype=np.int32)
        # good_dim_order
        newvar = ncfile.createVariable('good_dim_order', np.float32,
                                        dimensions=('time', 'lev', 'lat', 'lon', 'bnds'))
        newvar[:] = np.array(list(range(0,2**5)), dtype=np.float32).reshape(2,2,2,2,2)
        newvar.setncatts( { '_FillValue': np.array([300.0], dtype=np.float32),
                                        'valid_min': np.array([0.0], dtype=np.float32),
                                        'valid_max': np.array([200.0], dtype=np.float32),
                                        'history': 'this has always been a test variable',
                                        'institution': 'FZJ',
                                        'source': 'unit test helpers',
                                        'references': 'none',
                                        'comment': 'this variable has been created as a CF-1.0 compliant test variable',
                                        'long_name': 'good dimension order',
                                        'standard_name': 'wind_speed_of_gust number_of_observations',
                                        'units': '1',
                                        'ancillary_variables': 'measurements_status',
                                        'cell_measures': 'area: cell_areas',
                                        'cell_methods': 'time: mean lev: mean lon: lat: maximum '
                                                               '(interval: 0.2 degrees_E interval: 0.4 degrees_N)',
                                        })
        # cell_areas
        newvar = ncfile.createVariable('cell_areas', np.float32, dimensions=('lat', 'lon'))
        newvar[:] = np.array([[1.5,1.6],[1.5,1.5]], dtype=np.float32)
        newvar.setncatts( { 'long_name': 'cell areas',
                                        'units': 'inch2',
                                        })

        # measurements_status
        newvar = ncfile.createVariable('measurements_status', np.int8, dimensions=('time',))
        newvar[:] = np.array([0,2], dtype=np.int8)
        newvar.setncatts( { 'standard_name': 'wind_speed_of_gust status_flag',
                                        'flag_values': np.array([0,1,2], dtype=np.int8),
                                        'flag_meanings': 'good_data bad_data worse_data',
                                        'scale_factor': np.array(5, dtype=np.float64),
                                        })
        # measurements_status_2
        newvar = ncfile.createVariable('measurements_status_2', 'S1', dimensions=('time',))
        newvar[:] = _str_to_np_arr([chr(1)+chr(2)], len(ncfile.dimensions['time']))
        newvar.setncatts( { 'standard_name': 'wind_speed_of_gust status_flag',
                                        'flag_values': chr(1)+chr(2)+chr(3),
                                        'flag_meanings': 'good_data bad_data worse_data',
                                        })
        # is_alias_name
        newvar = ncfile.createVariable('is_alias_name', np.float32, dimensions=('lat','lon'))
        newvar[:] = np.array([[5,4],[3,2]], dtype=np.float32)
        newvar.setncatts( { 'standard_name': 'surface_temperature_where_land',
                                        'units': 'degree_C',
                                        })
        # T
        newvar = ncfile.createVariable('T', np.float32, dimensions=('lev', 'yc', 'xc'))
        newvar[:] = np.array(list(range(0,2**3)), dtype=np.float32).reshape(2,2,2)
        newvar.setncatts( { 'long_name': 'temperature',
                                        'standard_name': 'air_potential_temperature',
                                        'units': 'K',
                                        'coordinates': 'lon2d lat2d lev_name',
                                        })
        # ap
        newvar = ncfile.createVariable('ap', np.float32, dimensions=('lev',))
        newvar[:] = [500.,600.]
        newvar.setncatts( { 'long_name': 'atmosphere hybrid sigma a coordinate',
                                        'units': '1',
                                        })
        # ap2
        newvar = ncfile.createVariable('ap2', np.float32, dimensions=('lev2',))
        newvar[:] = [500.,600.]
        newvar.setncatts( { 'long_name': 'atmosphere hybrid sigma a coordinate',
                                        'units': '1',
                                        })
        # b
        newvar = ncfile.createVariable('b', np.float32, dimensions=('lev',))
        newvar[:] = [0.2,0.01]
        newvar.setncatts( { 'long_name': 'atmosphere hybrid sigma b coordinate',
                                        'units': '1',
                                        })
        # b2
        newvar = ncfile.createVariable('b2', np.float32, dimensions=('lev2',))
        newvar[:] = [0.2,0.01]
        newvar.setncatts( { 'long_name': 'atmosphere hybrid sigma b coordinate',
                                        'units': '1',
                                        })
        # p0
        newvar = ncfile.createVariable('p0', np.float32, dimensions=())
        newvar[:] = 1000
        newvar.setncatts( { 'long_name': 'reference pressure',
                                        'units': 'hPa',
                                        })
        # ps
        newvar = ncfile.createVariable('ps', np.float32, dimensions=('time', 'lat', 'lon'))
        newvar[:] = np.array([1024.0]* 2**3, dtype=np.float32).reshape(2,2,2)
        newvar.setncatts( { 'long_name': 'surface pressure',
                                        'units': 'hPa',
                                        'scale_factor': np.array(1.0, dtype=np.float32),
                                        'add_offset': np.array(0., dtype=np.float32),
                                        })
        # lev_name
        newvar = ncfile.createVariable('lev_name', 'S1', dimensions=('lev', 'max_label_len'))
        newvar[:] = _str_to_np_arr(['abcdef','ghijkl'], len(ncfile.dimensions['max_label_len']))

        # zero_sized_data_var
        newvar = ncfile.createVariable('zero_sized_data_var', np.float32, dimensions=('zero_sized_unlimited_dim', 'lat', 'lon'))
        newvar.setncatts({ 'long_name': 'zero-sized data variable',
                                        'units': '1',
                                        '_FillValue': np.array([np.nan], dtype=np.float32),
                                        'missing_value': np.array([np.nan], dtype=np.float32)})
    finally:
        ncfile.close()

def amend_faulty_ncfile_cf12(filename):
    '''
    additions:
    5.6: grid_mapping_4 has valid grid_mapping_name but unknown grid_* attributes
    '''
    ncfile = ncdf4(filename, mode='a', format='NETCDF3_CLASSIC')
    try:
        # modify grid_mapping_4 with new name
        newvar = ncfile.variables['grid_mapping_4']
        newvar.setncatts( { 'grid_mapping_name': 'latitude_longitude',
                                        'longitude_of_prime_meridian':  np.array(200.0,dtype=np.float32),
                                        })
    finally:
        ncfile.close()

def amend_faulty_ncfile_cf13(filename):
    '''
    additions:
    3.5:
    - _spam_data flag_masks attribute of different type than variable
    - _spam_data var type not compatible with bit field expressions (caused by flag_masks)
    - CaseTest flag_masks has different len than flag_meanings
    - double_axis flag_masks and flag_values bool AND are not equal flag_values
    '''
    ncfile = ncdf4(filename, mode='a', format='NETCDF3_CLASSIC')
    try:
        # modify _spam_data
        newvar = ncfile.variables['_spam_data']
        newvar.setncatts( { 'flag_masks': np.array([-9, 0, -7], dtype=np.int32),
                                        })
        # modify CaseTest
        newvar = ncfile.variables['CaseTest']
        newvar.setncatts( { 'flag_masks': np.array([-9, -8, -7, 12], dtype=np.int32),
                                        })
        # modify double_axis
        newvar = ncfile.variables['double_axis']
        newvar.setncatts( { 'flag_masks': np.array([1, 2, 3, 12, 16], dtype=np.int32),
                                        'flag_values': np.array([1, 2, 3, 4, 15], dtype=np.int32),
                                        })
    finally:
        ncfile.close()

def amend_faulty_ncfile_cf14(filename):
    '''
    additions:
    7.3:
    - wrong_dim_order: cell_methods contains lev, but lev has no bounds
    '''
    ncfile = ncdf4(filename, mode='a', format='NETCDF3_CLASSIC')
    try:
        var = ncfile.variables['wrong_dim_order']
        var.setncatts( { 'cell_methods': 'lon: sum lev: sum lat: variance '
                                                              '(interval: 5.0 comment: units missing in interval)',
                                        })

        var = ncfile.variables['double_dependency']
        var.setncatts( {  'cell_methods': 'time: sum where sea_foo over bla_foo within years time: mean over years',
                                        })
    finally:
        ncfile.close()

def amend_faulty_ncfile_cf15(filename):
    '''
    additions:
    3.5:
    - wrong_dim_order: flag_meanings phrases contain invalid chars,
                       already included in original cf10 test file
    '''
    ncfile = ncdf4(filename, mode='a', format='NETCDF3_CLASSIC')
    try:
        pass
    finally:
        ncfile.close()

def amend_faulty_ncfile_cf16(filename):
    '''
    additions:
    '''
    ncfile = ncdf4(filename, mode='a', format='NETCDF3_CLASSIC')
    try:
        pass
    finally:
        ncfile.close()

def amend_good_ncfile_cf11(filename):
    '''
    2.6.1: change Conventions attribute to CF-1.1
    '''
    ncfile = ncdf4(filename, mode='a', format='NETCDF3_CLASSIC')
    try:
        ncfile.setncatts( { 'Conventions': 'CF-1.1',
                                        })
    finally:
        ncfile.close()

def amend_good_ncfile_cf12(filename):
    '''
    2.6.1: change Conventions attribute to CF-1.2
    5.6: add new grid_mapping from CF-1.2
    '''
    ncfile = ncdf4(filename, mode='a', format='NETCDF3_CLASSIC')
    try:
        ncfile.setncatts( { 'Conventions': 'CF-1.2',
                                    })
        # add grid_mapping attribute to T
        ncfile.variables['T'].setncatts( { 'grid_mapping': 'crs',
                                                        })
        # create crs grid mapping variable
        newvar = ncfile.createVariable('crs', np.int32, dimensions=())
        newvar[:] = np.array([1],dtype=np.float32)
        newvar.setncatts( { 'grid_mapping_name': 'vertical_perspective',
                                        'latitude_of_projection_origin': np.array([10.0]),
                                        })
    finally:
        ncfile.close()

def amend_good_ncfile_cf13(filename):
    '''
    3.5:
    - measurements_status_2 has valid flag_masks attribute
    '''
    ncfile = ncdf4(filename, mode='a', format='NETCDF3_CLASSIC')
    try:
        ncfile.setncatts( { 'Conventions': 'CF-1.3',
                                    })
        # modify measurements_status_2
        var = ncfile.variables['measurements_status_2']
        var.setncatts( { 'flag_masks': chr(1)+chr(2)+chr(7),
                                  })
    finally:
        ncfile.close()

def amend_good_ncfile_cf14(filename):
    '''
    7.3:
    - good_dim_order with area in cell_methods
    - scalar_var has bounds variable
    - good_dim_order uses user-defined area type custom_area
    '''
    ncfile = ncdf4(filename, mode='a', format='NETCDF3_CLASSIC')
    try:
        # additional dimension for custom_area
        ncfile.createDimension('num_strings', 1)
        setattr(ncfile,'Conventions', 'CF-1.4')
        # create scalar coordinate variable for good_dim_order
        newvar = ncfile.createVariable('scalar_var', np.int32, dimensions=())
        newvar[:] = np.array(5)
        newvar.setncatts( { 'long_name': 'scalar coordinate variable',
                                        'units': '1',
                                        'bounds': 'scalar_bnds',
                                        })
        # create bounds variable
        newvar = ncfile.createVariable('scalar_bnds', np.int32, dimensions=('bnds',))
        newvar[:] = np.array([4,6], dtype=np.int32)
        # good_dim_order
        var = ncfile.variables['good_dim_order']
        var.setncatts( { 'cell_methods': 'time: mean lev: mean area: maximum scalar_var: sum where sea_ice over custom_area'
                                                        ' (interval: 0.2 degrees_E interval: 0.4 degrees_N)',
                                 'coordinates': 'scalar_var custom_area',
                                 })
        # custom_area
        newvar = ncfile.createVariable('custom_area', 'S1', dimensions=('num_strings', 'max_label_len'))
        newvar[:] = _str_to_np_arr(['foobar'], len(ncfile.dimensions['max_label_len']))
        newvar.setncatts( { 'standard_name': 'area_type',
                                        })
    finally:
        ncfile.close()

def amend_good_ncfile_cf15(filename):
    '''
    2.6.1: update conventions attribute to CF-1.5
    '''
    ncfile = ncdf4(filename, mode='a', format='NETCDF3_CLASSIC')
    try:
        setattr(ncfile,'Conventions', 'CF-1.5')
    finally:
        ncfile.close()

def amend_good_ncfile_cf16(filename):
    '''
    2.6.1: update conventions attribute to CF-1.6
    '''
    ncfile = ncdf4(filename, mode='a', format='NETCDF3_CLASSIC')
    try:
        setattr(ncfile,'Conventions', 'CF-1.6')
    finally:
        ncfile.close()
